#include "dual_quaternions.h"
#include <cstdlib>
#include <iomanip>
#include <sstream>
#include "mesh.h"

using std::vector;
using demo::matrix3x3;
using demo::point3d;

/**
 *
 */
void run_diffusion(
      int links_per_view,
      const vector< std::pair<matrix3x3<double>,point3d> >& pairwise,
      vector< std::pair<matrix3x3<double>,point3d> >& refined
){
   const int nranges = refined.size();
   std::cout << std::fixed << std::setprecision(8);

   transducer x(nranges);

   int p=0;
   for (int i=0; i<nranges; ++i)
   {
      const int P = i;

      int prev = P;
      for (int edge=0; edge<links_per_view; ++edge)
      {
         const int Q = (prev == nranges-1 ? 0 : prev+1);
         prev = Q;

         x.add_transformation(
               P, Q,
               demo::rot_matrix_to_quaternion( pairwise[p].first ),
               pairwise[p].second);
         ++p;
      }
   }

   x.get_estimate();
   std::cout << "Initial RMSTE " << x.rmste() << std::endl;

   x.transduce();
   std::cout << "Final RMSTE " << x.rmste() << std::endl;

   for (int i=0; i<nranges; ++i)
   {
      dual_quaternion tt;
      x.get_position(i,tt);

      matrix3x3<double> R = demo::quaternion_to_rot_matrix(tt.R);
      point3d T = tt.get_translation();

      refined[i].first = R;
      refined[i].second = T;
      demo::invert(refined[i].first, refined[i].second);
   }
}

template <typename T>
std::string ntos(T n)
{
   std::stringstream ss;
   ss << n;
   return ss.str();
}

void unhandled_exception()
{
   std::cerr << "Unhandled exception or terminate() caught." << std::endl;
#ifdef WIN32
   system("pause");
#endif
   abort();
}

int main(int argc, char** argv)
{
   if (argc != 2)
   {
      std::cout << "Usage: ./demo <data_dir>" << std::endl;
#ifdef WIN32
      system("pause");
#endif
      return 0;
   }

   std::set_terminate(unhandled_exception);
   std::cout << std::fixed << std::setprecision(8);

   const int nranges = 12;

   // Load all the rangemaps

   vector<mesh*> rangemaps(nranges);

   for (int i=0; i<nranges; ++i)
   {
      std::cout << "Loading rangemap " << i+1 << "/" << nranges << "... " << std::flush;

      try {
         rangemaps[i] = new mesh;
         rangemaps[i]->load_mesh_from_ply(
               std::string(argv[1]) + "/view" + ntos(i) + ".ply", false);
      } catch (std::exception& e) {
         std::cerr << "[ERROR] Exception caught: " << e.what() << std::endl;
      }

      std::cout << std::endl;
   }

   // Run Dual Quaternion Diffusion

   const int links_per_view = 3;

   // absolute motions
   vector< std::pair<matrix3x3<double>,point3d> > refined(nranges);
   for (int i=0; i<nranges; ++i) demo::set_identity(refined[i].first);

   // pairwise motions
   vector< std::pair<matrix3x3<double>,point3d> > pairwise;
   for (int i=0; i<nranges; ++i)
   {
      const int P = i;
      int prev = P;
      for (int step=0; step<links_per_view; ++step)
      {
         const int Q = (prev == nranges-1 ? 0 : prev+1);
         prev = Q;

         matrix3x3<double> R_PQ; point3d T_PQ;
         demo::relative_motion( refined[P].first, refined[P].second,
                                refined[Q].first, refined[Q].second,
                                R_PQ, T_PQ );

         pairwise.push_back( std::make_pair(R_PQ, T_PQ) );
      }
   }

   run_diffusion(links_per_view, pairwise, refined);

   // Save global alignment

   for (int i=0; i<nranges; ++i)
   {
      rangemaps[i]->apply_rigid_transform(refined[i].first, refined[i].second);

      vector<demo::color_rgb24> colors(rangemaps[i]->get_n_vertices(), demo::color_rgb24(rand()%255,rand()%255,rand()%255));
      rangemaps[i]->save_as_ply_colors(std::string(argv[1]) + "/view" + ntos(i) + "-aligned.ply", colors);
   }

   // Cleanup

   for (int i=0; i<nranges; ++i)
      delete rangemaps[i];

#ifdef WIN32
   system("pause");
#endif
   return 0;
}
